###############################################################################
# (c) Copyright 2020-2021 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""
Entry point functions for nightly build operations.
"""
import json
import os
import pathlib
import socket
from multiprocessing import Pool
from pathlib import Path
from subprocess import STDOUT, check_output, run
from time import sleep

import archspec.cpu
from lb.nightly.configuration import (
    DataProject,
    Package,
    Project,
    get,
    lbnightly_settings,
    service_config,
)
from lb.nightly.db import connect as db
from lb.nightly.utils import Repository

from . import CMAKE_DIR
from .common import download_and_unzip, get_build_method


def checkout(project_id: str, worker_task_id: str, gitlab_token=None):
    from .checkout import git, notify_gitlab, update_nightly_git_archive

    conf = service_config()

    gitlab_token = (
        gitlab_token
        or conf.get("gitlab", {}).get("token")
        or os.environ.get("GITLAB_TOKEN")
    )

    project = get(project_id)
    if not isinstance(project, (Project, Package)):
        raise ValueError(
            f"project_id {project_id} does not identify a nightly builds Project instance"
        )

    db().checkout_start(project, worker_task_id)

    # checkout the project
    report = git(project)

    # record the working directory
    report.cwd = os.getcwd()

    if gitlab_token:
        if isinstance(project, DataProject):
            for pkg, pkg_report in zip(project.packages, report.packages):
                notify_gitlab(pkg, pkg_report, gitlab_token)
        else:
            report = notify_gitlab(project, report, gitlab_token)

    # gitlab archive and dependencies are not possible for data projects
    if not isinstance(project, (DataProject, Package)):
        try:
            report = update_nightly_git_archive(project, report)
        except Exception as err:
            report.warning(f"failed to update Git archive: {err}")

        # resolve and update dependencies
        db().set_dependencies(project)

    if isinstance(project, Package):
        dir_to_pack = os.path.join(
            project.container.name, project.name, project.version
        )
    else:
        dir_to_pack = project.name
    archive_name = os.path.basename(project.artifacts("checkout"))

    report.info(f"packing {dir_to_pack} into {archive_name}")
    report.log(
        "stdout",
        "debug",
        check_output(
            [
                "zip",
                "-r",
                archive_name,
                dir_to_pack,
            ],
            stderr=STDOUT,
        ).decode(),
    )

    report.info(f"adding logs to {archive_name}")

    log_dir = pathlib.Path(project.name) / ".logs" / "checkout"
    os.makedirs(log_dir)
    with open(log_dir / "report.json", "w") as f:
        json.dump(report.to_dict(), f, indent=2)
    with open(log_dir / "report.md", "w") as f:
        f.write(report.md())

    report.log(
        "stdout",
        "debug",
        check_output(
            ["zip", "-r", archive_name, log_dir],
            stderr=STDOUT,
        ).decode(),
    )

    if "artifacts" in conf:
        artifacts_repo = Repository.connect(conf["artifacts"]["uri"])
        assert artifacts_repo.push(
            open(archive_name, "rb"), project.artifacts("checkout")
        ), "failed to upload artifacts"
    else:
        report.warning("artifacts repository not configured: no publishing")

    if "logs" in conf:
        logs_repo = Repository.connect(conf["logs"]["uri"])
        assert logs_repo.push(
            open(log_dir / "report.json", "rb"),
            project.artifacts("checkout") + "-report.json",
        ), "failed to upload log"

    else:
        report.warning("logs repository not configured: no publishing")

    db().checkout_complete(project, report, worker_task_id)

    return report


def build(
    project_id: str,
    platform: str,
    worker_task_id: str,
    gitlab_token=None,
):

    conf = service_config()
    repo = Repository.connect()

    project = get(project_id)
    if not isinstance(project, Project):
        raise ValueError(
            f"project_id {project_id} does not identify a nightly builds Project instance"
        )

    db().build_start(project, platform, worker_task_id)

    project_sources_dir = project.get_deployment_directory()
    attempts = 0
    while attempts < 10:
        if project_sources_dir.exists():
            break
        sleep(300)
        attempts += 1
    else:
        raise RuntimeError(
            f"cannot find project sources directory {project_sources_dir}"
        )

    build_env = {}
    build_env["BINARY_TAG"] = build_env["CMTCONFIG"] = platform
    build_env["PATH"] = os.environ["PATH"]
    build_env["LBENV_CURRENT_WORKSPACE"] = str(project.slot.get_deployment_directory())
    build_env[
        "CMAKE_PREFIX_PATH"
    ] = f"{build_env['LBENV_CURRENT_WORKSPACE']}:/cvmfs/lhcb.cern.ch/lib/lhcb:/cvmfs/lhcb.cern.ch/lib/lcg/releases:{CMAKE_DIR}:/cvmfs/projects.cern.ch/intelsw/psxe/linux/x86_64/2019/vtune_amplifier"

    if "TASK_LOGFILE" in os.environ:
        build_env["TASK_LOGFILE"] = os.environ["TASK_LOGFILE"]
    if "CONDA_ENV" in os.environ:
        build_env["CONDA_ENV"] = os.environ["CONDA_ENV"]
    build_env["NIGHTLIES_CACHE"] = lbnightly_settings().installations.path
    if "CMAKE_USE_CCACHE" not in project.slot.cache_entries:
        project.slot.cache_entries["CMAKE_USE_CCACHE"] = True

    report, build_logs = get_build_method(project)(project).build(
        jobs=int(os.environ.get("LBN_BUILD_JOBS") or "0") or os.cpu_count() or 1,
        init_env=build_env,
        worker_task_id=worker_task_id,
        cache_entries=project.slot.cache_entries,
    )
    archive_name = os.path.basename(project.artifacts("build", platform))
    report["script"].info(f"packing {project.name} into {archive_name}")
    report["script"].log(
        "stdout",
        "debug",
        check_output(
            [
                "zip",
                "-r",
                archive_name,
                project.name,
                "-i",
                f"{project.name}/build/*",
                f"{project.name}/InstallArea/*",
            ],
            stderr=STDOUT,
        ).decode(),
    )

    report["script"].info(f"adding logs to {archive_name}")

    log_dir = pathlib.Path(project.name) / ".logs" / platform / "build"
    os.makedirs(log_dir, exist_ok=True)
    with open(log_dir / "report.md", "w") as f:
        for step in report.values():
            f.write(step.md())
            f.write("\n")

    jreport = {}
    from .build_log import generate_build_report

    try:
        completed = report["install"].completed
    except AttributeError:
        try:
            completed = report["build"].completed
        except AttributeError:
            completed = report["configure"].completed
    retcodes = set()
    for rep in report.values():
        try:
            retcodes.add(rep.returncode)
        except AttributeError:
            pass

    reports_json = {k: v.to_dict() for k, v in report.items()}
    for val in reports_json.values():
        try:
            val["stdout"] = val["stdout"].decode(errors="surrogateescape")
            val["stderr"] = val["stderr"].decode(errors="surrogateescape")
        except KeyError:
            pass

    jreport = generate_build_report(
        build_logs=build_logs,
        proj_build_root=project.name,
        exceptions={
            "warning": project.slot.warning_exceptions,
            "error": project.slot.error_exceptions,
        },
        extra_info={
            "project": project.name,
            "version": project.version,
            "slot": project.slot.name,
            "slot_build_id": project.slot.build_id,
            "host": socket.gethostname(),
            "platform": platform,
            "started": report["configure"].started,
            "completed": completed,
            "retcode": max(retcodes),
            "environment": build_env,
            "cpu": archspec.cpu.detect.raw_info_dictionary(),
            "reports": reports_json,
        },
    )

    with open(log_dir / "report.json", "w") as f:
        json.dump(jreport, f, indent=2)

    report["script"].log(
        "stdout",
        "debug",
        check_output(["zip", "-r", archive_name, log_dir], stderr=STDOUT).decode(),
    )

    assert repo.push(
        open(archive_name, "rb"), project.artifacts("build", platform)
    ), "failed to upload artifacts"

    if "logs" in conf:
        logs_repo = Repository.connect(conf["logs"]["uri"])
        assert logs_repo.push(
            open(log_dir / "report.json", "rb"),
            project.artifacts("build", platform) + "-report.json",
        ), "failed to upload log"

    else:
        report["script"].warning("logs repository not configured: no publishing")

    db().build_complete(project, platform, jreport, worker_task_id)

    return report


def test(
    project_id: str,
    platform: str,
    worker_task_id: str = None,
    gitlab_token=None,
):

    conf = service_config()
    repo = Repository.connect()

    project = get(project_id)
    if not isinstance(project, Project):
        raise ValueError(
            f"project_id {project_id} does not identify a nightly builds "
            "Project instance"
        )

    db().tests_start(project, platform, worker_task_id)

    download_and_unzip(repo, project.artifacts("build", platform))

    build_env = {}
    build_env["BINARY_TAG"] = build_env["CMTCONFIG"] = platform
    build_env["PATH"] = os.environ["PATH"]
    build_env["LBENV_CURRENT_WORKSPACE"] = str(project.slot.get_deployment_directory())
    build_env[
        "CMAKE_PREFIX_PATH"
    ] = f"{build_env['LBENV_CURRENT_WORKSPACE']}:/cvmfs/lhcb.cern.ch/lib/lhcb:/cvmfs/lhcb.cern.ch/lib/lcg/releases:{CMAKE_DIR}:/cvmfs/projects.cern.ch/intelsw/psxe/linux/x86_64/2019/vtune_amplifier"
    if "TASK_LOGFILE" in os.environ:
        build_env["TASK_LOGFILE"] = os.environ["TASK_LOGFILE"]
    if "CONDA_ENV" in os.environ:
        build_env["CONDA_ENV"] = os.environ["CONDA_ENV"]
    build_env["NIGHTLIES_CACHE"] = lbnightly_settings().installations.path

    try:
        krb_auth = (
            lbnightly_settings()["kerberos"]["user"],
            lbnightly_settings()["kerberos"]["password"],
        )
    except KeyError:
        krb_auth = None

    try:
        os.remove(os.path.join(project.name, "build", "CMakeCache.txt"))
    except (OSError, FileNotFoundError):
        pass

    report = get_build_method(project)(project).test(
        init_env=build_env,
        krb_auth=krb_auth,
        worker_task_id=worker_task_id,
    )

    reports = {"report": report, "results": {}}

    from xml.etree import ElementTree as ET

    try:
        xml = ET.parse(next(pathlib.Path(".").glob("**/Test.xml")))
        status_translation = {
            "passed": "PASS",
            "failed": "FAIL",
            "skipped": "SKIPPED",
            "notrun": "SKIPPED",
            "error": "ERROR",
            "untested": "UNTESTED",
        }
        from itertools import groupby

        tests = sorted(
            xml.findall("./Testing/Test[@Status]"),
            key=lambda test: test.attrib["Status"],
        )
        summary = {
            status_translation[key]: sorted(test.find("Name").text for test in group)
            for key, group in groupby(tests, key=lambda test: test.attrib["Status"])
        }
        reports["results"] = summary
    except StopIteration:
        report["script"].error("failed to parse Test.xml (file is missing)")

    archive_name = os.path.basename(project.artifacts("test", platform))
    report["script"].info(f"packing {project.name} into {archive_name}")
    report["script"].info(f"adding logs to {archive_name}")

    log_dir = pathlib.Path(project.name) / ".logs" / platform / "test"
    os.makedirs(log_dir, exist_ok=True)
    for step in report.keys():
        with open(log_dir / "report.md", "a") as f:
            f.write(f"{report[step].md()}\n")

    reports_json = {k: v.to_dict() for k, v in report.items()}
    for val in reports_json.values():
        try:
            val["stdout"] = val["stdout"].decode(errors="surrogateescape")
            val["stderr"] = val["stderr"].decode(errors="surrogateescape")
        except KeyError:
            pass

    with open(log_dir / "report.json", "w") as f:
        json.dump(reports_json, f, indent=2)

    report["script"].log(
        "stdout",
        "debug",
        check_output(
            [
                "zip",
                "-r",
                archive_name,
                log_dir,
            ],
            stderr=STDOUT,
        ).decode(),
    )

    assert repo.push(
        open(archive_name, "rb"), project.artifacts("test", platform)
    ), "failed to upload artifacts"

    if "logs" in conf:
        logs_repo = Repository.connect(conf["logs"]["uri"])
        assert logs_repo.push(
            open(log_dir / "report.json", "rb"),
            project.artifacts("test", platform) + "-report.json",
        ), "failed to upload log"
        try:
            with open(f"{project.name}/build/Testing/TAG") as tagfile:
                assert logs_repo.push(
                    open(
                        f"{project.name}/build/Testing/{tagfile.readline().strip()}/Test.xml",
                        "rb",
                    ),
                    project.artifacts("build", platform) + "-Test.xml",
                ), "failed to upload Test.xml (pushing to repo failed)"
        except FileNotFoundError:
            report["script"].error("failed to upload Test.xml (file is missing)")

    else:
        report["script"].warning("logs repository not configured: no publishing")

    db().tests_complete(project, platform, reports, worker_task_id)

    return report


if __name__ == "__main__":  # pragma: no cover
    import logging
    import sys

    logging.basicConfig(level=logging.DEBUG)
    globals()[sys.argv[1]](*sys.argv[2:])



###############################################################################
# (c) Copyright 2020-2021 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import logging
import os
from datetime import datetime
from typing import Union

import gitlab
from lb.nightly.configuration import Package, Project


def get_gitlab_url(project: Union[Project, Package]):
    """
    Deduce the Gitlab Git URL for the given project.
    """
    # FIXME: the default URL should be retrieved from SoftConfDb
    if not isinstance(project, Package):
        group = "gaudi" if project.name == "Gaudi" else "lhcb"
    else:
        group = "lhcb-datapkg"
    return project.checkout_opts.get(
        "url", f"https://gitlab.cern.ch/{group}/{project.name}.git"
    )


def get_gitlab_id(url):
    """
    Deduce the Gitlab id for the given project URL.
    """
    from urllib.parse import urlparse

    gitlab_id = urlparse(url).path.lstrip("/")
    if gitlab_id.endswith(".git"):
        gitlab_id = gitlab_id[:-4]

    return gitlab_id


def postToMergeRequest(
    name_or_id, mreq_iid, message, new_comment=False, token=None, logger=logging
):
    """
    Add the passed message as comment to a merge request in gitlab.cern.ch.

    @param name_or_id: qualified name or id of a project in gitlab
    @param mreq_iid: local id of the merge request
    @param message: what to post to the merge request
    @param new_comment: whether to always post a new comment or edit
    @param token: gitlab API token (default: os.environ['GITLAB_TOKEN'])
    """
    server = gitlab.Gitlab(
        "https://gitlab.cern.ch/", token or os.environ["GITLAB_TOKEN"]
    )
    logger.debug("looking for merge request %s in project %s", mreq_iid, name_or_id)
    try:
        project = server.projects.get(name_or_id)
        mreq = project.mergerequests.get(mreq_iid)
        if new_comment:
            mreq.notes.create({"body": message})
        else:
            time_tag = datetime.now().strftime("%Y-%m-%d %H:%M")
            message = "- __[{}]__ {}".format(time_tag, message)
            server.auth()  # sets server.user to the authenticated user
            all_notes = mreq.notes.list(order_by="created_at", sort="desc")
            # NB: replying to a plain note makes it a discussion, so next
            # time it won't be found and a new one will be made.
            own_plain_notes = [
                note
                for note in all_notes
                if (
                    note.author["id"] == server.user.id
                    and not note.system
                    and note.type != "DiscussionNote"
                )
            ]
            if own_plain_notes:
                last_note = own_plain_notes[0]
                last_note.body = last_note.body + "\n" + message
                last_note.save()
            else:
                mreq.notes.create({"body": message})
    except Exception as err:
        logger.error(str(err))
        raise


def getMRTitle(name_or_id, mreq_iid, token=None, logger=logging):
    """
    Return the title of a merge request in gitlab.cern.ch.

    @param name_or_id: qualified name or id of a project in gitlab
    @param mreq_iid: local id of the merge request
    @param token: gitlab API token (default: os.environ['GITLAB_TOKEN'])
    """
    token = token or os.environ.get("GITLAB_TOKEN")
    if not token:
        # the MR title is icing on the cake, we do not need to fail
        return ""

    server = gitlab.Gitlab("https://gitlab.cern.ch/", token)
    logger.debug("looking for merge request %s in project %s", mreq_iid, name_or_id)
    try:
        project = server.projects.get(name_or_id)
        mreq = project.mergerequests.get(mreq_iid)
        return mreq.title
    except Exception as err:
        logger.error(str(err))
        raise


def gitlabProjectExists(project, logger=logging):
    """
    Quick check to see if a project exists and is public in Gitlab.
    """
    from urllib.request import urlopen

    try:
        logger.debug("probing %s in Gitlab", project)
        url = (
            "https://gitlab.cern.ch/{}.git/info/refs?service=git-upload-pack"
        ).format(project)
        return urlopen(url).getcode() == 200
    except Exception:
        return False


MR_COMMENT_TMPLS = {
    True: (
        "Validation started with [{slot}#{id}]("
        "https://lhcb-nightlies.web.cern.ch/nightly/{slot}/{id}/)"
    ),
    False: (
        "Automatic merge failed in [{slot}#{id}]("
        "https://lhcb-nightlies.web.cern.ch/nightly/{slot}/{id}/{proj}/checkout"
    ),
}


def notifyMergeRequest(
    proj: Project, name_or_id, mreq_iid, success, token=None, logger=logging
):
    """
    Post the link to the slot build as comment to a merge request.
    """
    if (
        not proj
        or not proj.slot
        or not proj.slot.build_id
        or (os.environ.get("NO_UPDATE_MR", "false").lower() not in ("false", "0", ""))
    ):
        # noting to notify
        return
    if not token and "GITLAB_TOKEN" not in os.environ:
        logger.warning(
            "cannot post comment to gitlab for project %s, mr %s", name_or_id, mreq_iid
        )
        return

    message = MR_COMMENT_TMPLS[success].format(
        slot=proj.slot.name, id=proj.slot.build_id, proj=proj.name
    )
    postToMergeRequest(name_or_id, mreq_iid, message, token=token)



###############################################################################
# (c) Copyright 2020-2021 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import re


class ClassifyByLineNo:
    """
    Helper to set specific class for groups of lines in TableizeLine.
    """

    def __init__(self, range_classes):
        self.range_classes = list(range_classes)

    def __call__(self, line_no):
        for (begin, end), css in self.range_classes:
            if line_no >= begin and line_no < end:
                return css
        return None


# cached regular expression to find ANSI color codes
COLCODE_RE = re.compile("\x1b\\[([0-9;]*m|[012]?K)")


class ANSIStyle:
    __slots__ = ("style", "color", "bgcolor")

    def __init__(self, style=0, color=0, bgcolor=0):
        self.style = style
        self.color = color
        self.bgcolor = bgcolor
        if isinstance(style, str):
            self.style = 0
            self.apply_code(style)

    def apply_code(self, code):
        """
        >>> ANSIStyle(1, 2, 3).apply_code('0')
        ANSIStyle(style=0, color=0, bgcolor=0)
        >>> ANSIStyle().apply_code('1;34')
        ANSIStyle(style=1, color=4, bgcolor=0)
        >>> ANSIStyle().apply_code('45')
        ANSIStyle(style=0, color=0, bgcolor=5)
        """
        if not code or code == "0":
            self.style = self.color = self.bgcolor = 0
        else:
            for subcode in [int(x, 10) for x in code.split(";")]:
                if subcode >= 40:
                    self.bgcolor = subcode - 40
                elif subcode >= 30:
                    self.color = subcode - 30
                else:
                    self.style = subcode
        return self

    def copy(self):
        return ANSIStyle(self.style, self.color, self.bgcolor)

    def code(self, base=None):
        if (self.style, self.color, self.bgcolor) == (0, 0, 0):
            return ""
        if not base:
            base = ANSIStyle()
        if self == base:
            return self.code()  # prevent a no change to become a reset
        codes = []
        if self.style != base.style:
            codes.append(str(self.style))
        if self.color != base.color:
            codes.append(str(30 + self.color))
        if self.bgcolor != base.bgcolor:
            codes.append(str(40 + self.bgcolor))
        return ";".join(codes)

    def css(self, base=None):
        """
        CSS class(es) for the current text style.

        >>> ANSIStyle().css()
        ''
        >>> ANSIStyle('1;32;43').css()
        'xterm-style-1 xterm-color-2 xterm-bgcolor-3'
        """
        if (self.style, self.color, self.bgcolor) == (0, 0, 0):
            return ""
        if not base:
            base = ANSIStyle()
        codes = []
        for key in self.__slots__:
            if getattr(self, key) != getattr(base, key):
                codes.append("xterm-{}-{}".format(key, getattr(self, key)))
        return " ".join(codes)

    def __repr__(self):
        return "ANSIStyle({})".format(
            ", ".join(
                "{}={!r}".format(key, getattr(self, key)) for key in self.__slots__
            )
        )

    def __eq__(self, other):
        return (
            isinstance(
                other,
                ANSIStyle,
            )
            and (self.style, self.color, self.bgcolor)
            == (other.style, other.color, other.bgcolor)
        )


class ANSI2HTML:
    """
    Class to convert ANSI codes into HTML classes.
    """

    def __init__(self, start_style=""):
        self.current_style = (
            start_style
            if isinstance(start_style, ANSIStyle)
            else ANSIStyle(start_style)
        )

    def _process(self, line):
        # we record and strip the newline format at end of line not to include
        # it in the formatting
        if line.endswith("\n"):
            newline = "\n"
            line = line[:-1]
        elif line.endswith("\r\n"):
            newline = "\r\n"
            line = line[:-2]
        else:
            newline = ""

        # special handling of styles at beginning of line
        m = COLCODE_RE.match(line)
        while m:
            line = line[m.end() :]
            if m.group(1).endswith("m"):
                self.current_style.apply_code(m.group(1)[:-1])
            m = COLCODE_RE.match(line)

        current_class = self.current_style.css()
        if current_class:
            yield '<span class="{}">'.format(current_class)
        else:
            yield "<span>"

        pos = 0
        while True:
            # look for a control sequence
            m = COLCODE_RE.search(line, pos)
            if not m:
                # no more codes to convert
                break
            # pass the chars so far
            if m.start() != pos:
                yield line[pos : m.start()]
            # parse the new code
            if m.group(1).endswith("m"):
                self.current_style.apply_code(m.group(1)[:-1])
                next_class = self.current_style.css()
            else:
                next_class = current_class

            if next_class != current_class:
                # we have a change, close previous span, open the new one
                yield '</span><span class="{}">'.format(next_class)
                current_class = next_class
            pos = m.end()  # update offset to end of control code

        if pos < len(line):  # flush what remains of the line
            yield line[pos:]
        yield "</span>" + newline  # close the global <span>

    def __call__(self, line):
        """
        Add HTML span tags for ansi colors.
        """
        return "".join(self._process(line))


class TableizeLine:
    """
    Add table row tags and optionally line numbers to lines.
    """

    def __init__(
        self, first_line_no=1, line_id_prefix="L", add_line_nos=False, row_class=None
    ):
        self.line_no = first_line_no

        line_format = '<tr id="{prefix}{{n}}"{{class_desc}}><td>'
        if add_line_nos:
            line_format += '<a href="#{prefix}{{n}}">{{n}}</a></td><td>'
        line_format += "{{text}}</td></tr>"

        if row_class is None or isinstance(row_class, str):
            self.row_class = lambda n: row_class
        else:
            self.row_class = row_class

        self._line_format = line_format.format(prefix=line_id_prefix)

    def __call__(self, line):
        """
        Add HTML tags.
        """
        # we record and strip the newline format at end of line not to include
        # it in the formatting
        if line.endswith("\n"):
            newline = "\n"
            line = line[:-1]
        elif line.endswith("\r\n"):
            newline = "\r\n"
            line = line[:-2]
        else:
            newline = ""

        css = self.row_class(self.line_no)
        class_desc = ' class="{}"'.format(css) if css else ""

        out = (
            self._line_format.format(n=self.line_no, class_desc=class_desc, text=line)
            + newline
        )
        self.line_no += 1
        return out


class WrapLine:
    """
    Wrap a line in HTML tags.
    """

    def __init__(self, tag, attrs={}):
        self._line_format = "<{tag}{attrs}>{{text}}</{tag}>".format(
            tag=tag,
            attrs=""
            if not attrs
            else " "
            + " ".join('{0}="{1}"'.format(item) for item in list(attrs.items())),
        )

    def __call__(self, line):
        """
        Add HTML tags.
        """
        # we record and strip the newline format at end of line not to include
        # it in the formatting
        if line.endswith("\n"):
            newline = "\n"
            line = line[:-1]
        elif line.endswith("\r\n"):
            newline = "\r\n"
            line = line[:-2]
        else:
            newline = ""
        return self._line_format.format(text=line) + newline



###############################################################################
# (c) Copyright 2020-2021 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import asyncio
import json
import logging
import os
import socket
from collections import namedtuple
from contextlib import redirect_stderr
from datetime import datetime, timezone
from io import StringIO
from pathlib import Path
from shutil import copyfileobj
from ssl import create_default_context
from subprocess import CalledProcessError, run
from tempfile import NamedTemporaryFile
from typing import Optional, Union
from zlib import decompress

from git import GitCommandError
from lb.nightly.configuration import DataProject, Package, Project

from . import DATA_DIR


class Report:
    """
    Class used to collect reports.

    Record messages and forward them to a logger.

    Apart from normal logging levels, accept extra type of messages, like
    ``git_error``.
    """

    def __init__(self, logger=__name__):
        self.logger = logging.getLogger(logger)
        self.records = []

    def log(self, type, level, fmt, args=None):
        record = {"type": type, "level": level, "text": (fmt % args if args else fmt)}
        if type in ("error", "warning") and f"{type}:" not in record["text"]:
            record["text"] = "{}: {}".format(type, record["text"])
        self.records.append(record)
        getattr(self.logger, level)(fmt, *(() if args is None else args))

    def debug(self, fmt, *args):
        self.log("debug", "debug", fmt, args)

    def info(self, fmt, *args):
        self.log("info", "info", fmt, args)

    def warning(self, fmt, *args):
        self.log("warning", "warning", fmt, args)

    warn = warning

    def error(self, fmt, *args):
        self.log("error", "error", fmt, args)

    def git_error(self, msg, err: GitCommandError):
        """
        Special handling for GitCommandError
        """
        self.warning("%s: GitCommand status %s", msg, err.status)
        # FIXME: this is very much subject to the internal details of GitCommandError
        self.log("command", "debug", "> %s", (err._cmdline,))
        if "'" in err.stdout:
            self.log(
                "stdout", "debug", err.stdout[err.stdout.index("'") + 1 : -1].rstrip()
            )
        if "'" in err.stderr:
            self.log(
                "stderr", "debug", err.stderr[err.stderr.index("'") + 1 : -1].rstrip()
            )

    def md(self):
        """
        Format as markdown.
        """
        out = []
        if hasattr(self, "project"):
            out.append("# {name}/{version}".format(**self.project))
        for r in self.records:
            lines = r["text"].splitlines()
            if r["type"] in ("stdout", "stderr", "command"):
                out.append("  ```")
                out.extend("  " + l for l in lines)
                out.append("  ```")
            else:
                out.append("- " + lines.pop(0))
                out.extend("  " + l for l in lines)
        if hasattr(self, "packages"):
            out.extend(pkg.md() for pkg in self.packages)
        return "\n".join(out)

    def __str__(self):
        out = []
        if hasattr(self, "project"):
            out.append("{name}/{version}".format(**self.project))
            out.append("-" * len(out[-1]))
            out.append("")
        out.extend(r["text"] for r in self.records)
        if hasattr(self, "packages"):
            out.extend(str(pkg) for pkg in self.packages)
        return "\n".join(out)

    def to_dict(self):
        from copy import deepcopy

        data = deepcopy(self.__dict__)
        data["logger"] = self.logger.name
        if "packages" in data:
            data["packages"] = [pkg.to_dict() for pkg in data["packages"]]

        return data


def ensure_dir(path, rep: Report):
    """
    Make sure that a directory exist, creating it
    and passing a warning to Report object if dir does not exist.
    """
    if not os.path.exists(path):
        rep.warning(f"directory {path} is missing, I create it")
        os.makedirs(path)


def find_path(name, search_path=None):
    """
    Look for a file or directory in a search path.

    If the search path is not specified, the concatenation of CMTPROJECTPATH
    and CMAKE_PREFIX_PATH is used.

    >>> find_path('true', ['/usr/local/bin', '/bin'])
    '/bin/true'
    >>> print(find_path('cannot_find_me', []))
    None
    """
    from os import environ, pathsep
    from os.path import exists, join

    if search_path is None:
        search_path = environ.get("CMAKE_PREFIX_PATH", "").split(pathsep) + environ.get(
            "CMTPROJECTPATH", ""
        ).split(pathsep)

    try:
        return next(
            join(path, name) for path in search_path if exists(join(path, name))
        )
    except StopIteration:
        logging.warning("%s not found in %r", name, search_path)
    return None


def get_build_method(project=None):
    """
    Helper function to get a build method for a project.

    The method is looked up in the following order: build_tool property of a project,
    build_tool property of a slot owning the project, the default build method (cmake).

    If a method is retrieved via build_tool string property, it must be defined in
    lb.nightly.functions.build.
    """
    import lb.nightly.functions.build as build_methods

    try:
        method = project.build_tool or project.slot.build_tool
        return getattr(build_methods, method)
    except (AttributeError, TypeError):
        return build_methods.cmake


def safe_dict(mydict):
    """Helper to return the dictionary without sensitive data
    To be used e.g. to remove secret environment variables.
    >>> d={"PASSWORD": "my_secret_pass", "PASS": "asd", "USER": "me", "KEY": "asd", "TOKEN": "asd", "PRIVATE": "Asd", "HOME": "Asd", "SHA": "asd"}
    >>> safe_dict(d)
    {}
    """
    return {
        k: v
        for k, v in mydict.items()
        if all(
            s not in k.upper()
            for s in [
                "PASS",
                "USER",
                "KEY",
                "TOKEN",
                "PRIVATE",
                "HOME",
                "SHA",
            ]
        )
    }


def singularity_run(
    cmd: list,
    env: Optional[dict] = None,
    cwd: str = os.curdir,
    task: Optional[dict] = None,
    krb_auth: Optional[tuple] = None,
):
    from spython.main import Client

    try:
        Client.load(get_singularity_root(binary_tag=env["BINARY_TAG"]))
    except AttributeError:
        raise RuntimeError(
            f"Could not find the CernVM image for the "
            f"{env.get('BINARY_TAG', 'not specified BINARY_TAG')}"
        )

    s_env = safe_dict(env)

    # specify workspace in container
    workspace_singularity = pwd = "/workspace"
    if cwd != os.getcwd():
        cwd = os.path.join(workspace_singularity, cwd)
        pwd = cwd

    if krb_auth:
        krb_token = NamedTemporaryFile()
        s_env["KRB5CCNAME"] = f"FILE:{krb_token.name}"

        class KerberosAuthenticationError(Exception):
            pass

        try:
            run(
                ["kinit", "-c", krb_token.name, krb_auth[0]],
                input=krb_auth[1].encode(),
                check=True,
            )
        except (CalledProcessError, IndexError):
            raise KerberosAuthenticationError(
                "Could not authenticate with provided credentials"
            )

    # setting up opensearch where we send logs from singularity
    from lb.nightly.configuration import service_config
    from opensearchpy import OpenSearch

    conf = service_config()
    es_url = conf.get("opensearch", {}).get("uri", "dummy_opensearch_uri")
    capath = conf.get("opensearch", {}).get("capath")
    index_pattern_prefix = conf.get("opensearch", {}).get(
        "index_pattern_prefix",
        "nightlies",
    )
    opensearch = OpenSearch(
        [es_url],
        ssl_context=create_default_context(capath=capath),
    )
    logging.getLogger("opensearch").setLevel(logging.WARNING)
    logging.getLogger("urllib3").setLevel(logging.WARNING)
    if opensearch.ping():
        # stub of the msg to send to opensearch
        log_body = {
            "host": socket.gethostname(),
            "command": cmd,
            "cwd": cwd,
            "task": task.get("task"),
            "project": task.get("project"),
            "platform": task.get("platform"),
            "worker_task_id": task.get("worker_task_id"),
            "log": None,  # will be overriden just before sending
            "timestamp": None,  # same as above
        }
        log_body.update(
            {
                "log": f"Running command: {cmd} with env: {s_env}",
                "timestamp": datetime.now(timezone.utc),
            }
        )
        opensearch.index(
            index=f"{index_pattern_prefix}-{datetime.now(timezone.utc).strftime('%Y.%m.%d')}-{task.get('task', 'nightlies')}",
            body=log_body,
        )

    if "TASK_LOGFILE" in env:
        with open(env.get("TASK_LOGFILE"), "a") as logfile:
            logfile.write(
                f"{datetime.now(timezone.utc).strftime('%Y-%m-%d %H:%M:%S')}: "
                f"Running command: {cmd} with {s_env}\n"
            )

    # to collect the build messages we communicate through the UNIX socket
    from hashlib import sha256

    hash = sha256()
    hash.update(f"{cmd} {task} {cwd}".encode())
    os.environ["LB_WRAPCMD_SOCKET"] = s_env["LB_WRAPCMD_SOCKET"] = os.path.join(
        "/tmp", hash.hexdigest()[:32]
    )

    Result = namedtuple(
        "Result",
        [
            "returncode",
            "stdout",
            "stderr",
            "args",
            "build_logs",
        ],
    )

    try:
        result = Client.execute(
            cmd,
            bind=[
                "/cvmfs",
                env.get(
                    "CONDA_ENV"
                ),  # safe because spython ignores "false" values in the list
                env.get("NIGHTLIES_CACHE"),
                f"{os.getcwd()}:{workspace_singularity}",
            ],
            options=[f"--pwd={pwd}", "--cleanenv"]
            + [f"--env={k}={v}" for k, v in s_env.items()],
            stream=True,
        )

        stdout = b""
        build_logs = {}

        async def get_stdout(result):
            nonlocal stdout
            for line in result:
                stdout += line.encode(errors="surrogateescape")
                yield line
                await asyncio.sleep(0)

        async def log_to_file(result):
            with open(
                os.environ.get("TASK_LOGFILE"),
                "a",
                buffering=512,
            ) as logfile:
                async for line in result:
                    logfile.write(
                        f"{datetime.now(timezone.utc).strftime('%Y-%m-%d %H:%M:%S')}: "
                        f"{line.rstrip()}\n"
                    )
                    yield line
                    await asyncio.sleep(0)

        async def log_to_opensearch(result):
            async for line in result:
                yield dict(
                    log_body,
                    **{
                        "_index": f"{index_pattern_prefix}-{datetime.now(timezone.utc).strftime('%Y.%m.%d')}-{task.get('task', 'nightlies')}",
                        "log": line.rstrip(),
                        "timestamp": datetime.now(timezone.utc),
                    },
                )
                await asyncio.sleep(0)

        from opensearchpy import AsyncOpenSearch
        from opensearchpy.helpers import async_streaming_bulk

        aopensearch = AsyncOpenSearch(
            hosts=[es_url],
            ssl_context=create_default_context(capath=capath),
        )

        async def stream_to_opensearch(result):
            async for _ in async_streaming_bulk(
                aopensearch,
                result,
                chunk_size=100,
                raise_on_exception=False,
                raise_on_error=False,
            ):
                await asyncio.sleep(0)

        async def stream_to_file(result):
            async for _ in log_to_file(result):
                await asyncio.sleep(0)

        if "TASK_LOGFILE" in env:

            async def get_output_and_log(result):
                if opensearch.ping():
                    await stream_to_opensearch(
                        log_to_opensearch(log_to_file(get_stdout(result)))
                    )
                else:
                    await stream_to_file(get_stdout(result))

        else:

            async def get_output_and_log(result):
                if opensearch.ping():
                    await stream_to_opensearch(log_to_opensearch(get_stdout(result)))
                else:
                    async for _ in get_stdout(result):
                        await asyncio.sleep(0)

        async def read_from_singularity(result):
            nonlocal stdout
            await get_output_and_log(result)
            return stdout

        server_address = os.environ["LB_WRAPCMD_SOCKET"]
        try:
            os.unlink(server_address)
        except OSError:
            if os.path.exists(server_address):
                raise

        async def _receive(reader, writer):
            buffer = b""
            while True:
                msg = await reader.read(4096)
                if msg:
                    buffer += msg
                else:
                    break
            data = json.loads(decompress(buffer))
            build_logs.setdefault(data.pop("subdir"), {}).setdefault(
                data.pop("target"), []
            ).append(data)

        async def read_socket(server_address):
            server = await asyncio.start_unix_server(
                _receive,
                server_address,
            )
            async with server:
                try:
                    await server.serve_forever()
                except asyncio.exceptions.CancelledError:
                    pass

        async def main_get_read(result, server_address):
            nonlocal stdout
            asyncio.create_task(read_socket(server_address))
            task = asyncio.create_task(read_from_singularity(result))
            try:
                await task
            except asyncio.exceptions.CancelledError:
                pass
            return stdout

        with redirect_stderr(StringIO()) as singularity_stderr:
            stdout = asyncio.run(main_get_read(result, server_address))

        # will be overwritten if CalledProcessError is raised
        retcode = 0

    except CalledProcessError as cpe:
        retcode = cpe.returncode
    finally:
        try:
            os.unlink(server_address)
        except OSError:
            pass
        try:
            del os.environ["LB_WRAPCMD_SOCKET"]
        except KeyError:
            pass
        try:
            krb_token.close()
        except NameError:
            pass
    return Result(
        stdout=stdout,
        returncode=retcode,
        stderr=singularity_stderr.getvalue().encode(errors="surrogateescape"),
        args=cmd,
        build_logs=build_logs,
    )


async def write_to_socket(socket_name, message):
    _, writer = await asyncio.open_unix_connection(socket_name)
    writer.write(message)
    writer.close()


def lb_wrapcmd():
    """
    The script to be used in CMake launcher rules.

    A launcher rule defined as
    set_property(GLOBAL PROPERTY RULE_LAUNCH_COMPILE "lb-wrapcmd <CMAKE_CURRENT_BINARY_DIR> <TARGET_NAME>")

    Output for each compile command is sent to the socket
    (name taken from LB_WRAPCMD_SOCKET env variable or built as:
    <CMAKE_CURRENT_BINARY_DIR>/1234-<TARGET_NAME>-abc-build.log)
    """
    import sys
    from argparse import REMAINDER, ArgumentParser
    from shutil import which
    from time import time
    from zlib import compress

    parser = ArgumentParser()
    parser.add_argument("log_dir", help="CMAKE_CURRENT_BINARY_DIR")
    parser.add_argument("target", help="TARGET_NAME")
    parser.add_argument("wrap_cmd", nargs=REMAINDER, help="wrapped command")
    args = parser.parse_args()

    with NamedTemporaryFile() as statistics_file:
        time_cmd = [
            which("time"),
            "-f",
            '{"duration": "%E", "maxresident": "%M", "avgresident": "%t", "cpupercentage": "%P"}',
            "-o",
            statistics_file.name,
        ]
        started = time()
        result = run(time_cmd + args.wrap_cmd, capture_output=True)
        completed = time()
        sys.stdout.buffer.write(result.stdout)
        sys.stderr.buffer.write(result.stderr)

        if "LB_WRAPCMD_SOCKET" in os.environ:
            server_address = os.environ["LB_WRAPCMD_SOCKET"]

            output = {
                "subdir": "/".join(args.log_dir.split("/")[4:]),
                "target": args.target,
                "cmd": args.wrap_cmd,
                "stdout": "",
                "stderr": "",
                "returncode": result.returncode,
                "started": started,
                "completed": completed,
            }
            statistics_file.seek(0)
            with open(statistics_file.name, "r") as f:
                statistics = json.loads(f.read())
            output.update(statistics)

            if args.wrap_cmd[0] != "cat":
                output["stdout"] = result.stdout.decode(errors="surrogateescape")
                output["stderr"] = result.stderr.decode(errors="surrogateescape")

            asyncio.run(
                write_to_socket(server_address, compress(json.dumps(output).encode()))
            )

    return result.returncode


def download_and_unzip(repo, artifact):
    """
    Helper function to download the artifact
    and unzip it in the current directory.

    :param repo: ArtifactsRepository object to get the artifact from
    :param artifact: the value returned by the `artifacts`
    method of the Project.

    Returns True if succeded.
    """
    try:
        with NamedTemporaryFile() as lp:
            with repo.pull(artifact) as remote:
                copyfileobj(remote, lp)
            lp.seek(0)
            return not run(["unzip", "-o", lp.name], check=True).returncode
    except Exception:
        return False


def create_project_makefile(dest, overwrite=False):
    """Write the generic Makefile for CMT projects.
    @param dest: the name of the destination file
    @param overwrite: flag to decide if an already present file has to be kept
                      or not (default is False)
    """
    import logging

    if overwrite or not os.path.exists(dest):
        logging.debug("Creating '%s'", dest)
        with open(dest, "w") as f:
            f.write(
                "DEVTOOLS_DATADIR := {0}\n"
                "include $(DEVTOOLS_DATADIR)/Makefile-common.mk\n".format(DATA_DIR)
            )
        return True
    return False


def create_toolchain_file(dest, overwrite=False):
    """Write the generic toolchain.cmake file needed by CMake-based projects.
    @param dest: destination filename
    @param overwrite: flag to decide if an already present file has to be kept
                      or not (default is False)
    """
    import logging

    if overwrite or not os.path.exists(dest):
        logging.debug("Creating '%s'", dest)
        with open(dest, "w") as f:
            f.write("include({0})\n".format(os.path.join(DATA_DIR, "toolchain.cmake")))
        return True
    return False


def create_git_ignore(dest, overwrite=False, extra=None, selfignore=True):
    """Write a generic .gitignore file, useful for git repositories.
    @param dest: destination filename
    @param overwrite: flag to decide if an already present file has to be kept
                      or not (default is False)
    @param extra: list of extra patterns to add
    @param selfignore: if the .gitignore should include itself
    """
    import logging

    if overwrite or not os.path.exists(dest):
        logging.debug("Creating '%s'", dest)
        patterns = [
            "/InstallArea/",
            "/build.*/",
            "*.pyc",
            "*~",
            ".*.swp",
            "/.clang-format",
        ]
        if selfignore:
            patterns.insert(0, "/.gitignore")  # I like it as first entry
        if extra:
            patterns.extend(extra)

        with open(dest, "w") as f:
            f.write("\n".join(patterns))
            f.write("\n")
        return True
    return False


def create_clang_format(dest, overwrite=False):
    """Add `.clang-format` file.
    @param dest: destination filename
    @param overwrite: flag to decide if an already present file has to be kept
                      or not (default is False)
    """
    import logging

    if overwrite or not os.path.exists(dest):
        logging.debug("Creating '%s'", dest)
        with open(dest, "w") as f:
            f.writelines(open(os.path.join(DATA_DIR, "default.clang-format")))
        return True
    return False


def init_project(path, overwrite=False):
    """
    Initialize the sources for an LHCb project for building.

    Create the (generic) special files required for building LHCb/Gaudi
    projects.

    @param path: path to the root directory of the project
    @param overwrite: whether existing files should be overwritten, set it to
                      True to overwrite all of them or to a list of filenames
    """
    extraignore = []
    factories = [
        ("Makefile", create_project_makefile),
        ("toolchain.cmake", create_toolchain_file),
        (
            ".gitignore",
            lambda dest, overwrite: create_git_ignore(dest, overwrite, extraignore),
        ),
        (".clang-format", create_clang_format),
    ]

    # handle the possible values of overwrite to always have a set of names
    if overwrite in (False, None):
        overwrite = set()
    elif overwrite is True:
        overwrite = set(f[0] for f in factories)
    elif isinstance(overwrite, str):
        overwrite = set([overwrite])
    else:
        overwrite = set(overwrite)

    for filename, factory in factories:
        if factory(os.path.join(path, filename), overwrite=filename in overwrite):
            extraignore.append("/" + filename)


def get_singularity_root(binary_tag: str):
    """
    >>> get_singularity_root("x86_64-centos7-gcc9-opt")
    '/cvmfs/cernvm-prod.cern.ch/cvm4'
    >>> get_singularity_root("x86_64-slc6-gcc9-opt")
    '/cvmfs/cernvm-prod.cern.ch/cvm3'
    >>> get_singularity_root("x86_64-slc5-gcc9-opt")
    '/cvmfs/cernvm-prod.cern.ch/cvm3'
    >>> get_singularity_root("x86_64-ubuntu-gcc9-opt")
    >>>
    """
    from LbPlatformUtils import OS_ALIASES, OS_COMPATIBILITY
    from LbPlatformUtils.inspect import SINGULARITY_ROOTS

    wanted_os = binary_tag.split("-")[1]
    compatible_oss = [wanted_os] + [
        main_os
        for main_os, comp_oss in OS_COMPATIBILITY.items()
        if wanted_os in comp_oss
    ]
    try:
        compatible_aliases = [
            comp_os
            for compatible_os in compatible_oss
            for comp_os in OS_ALIASES[compatible_os]
        ]
        return [path for path, osid in SINGULARITY_ROOTS if osid in compatible_aliases][
            0
        ]
    except (IndexError, KeyError):
        return None



###############################################################################
# (c) Copyright 2020-2021 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import os
import re
from datetime import datetime
from pathlib import Path
from shlex import quote

from lb.nightly.configuration import Project

from .common import Report, ensure_dir, find_path, init_project, singularity_run


class cmake_new:

    # known "kwargs" not to be propagated to subprocess
    SPECIAL_ARGS = ("jobs", "args", "make_cmd", "cache_entries", "step")

    def __init__(self, project: Project) -> None:
        self.project = project
        if not isinstance(self.project, Project):
            raise TypeError(
                f"expected lb.nightly.configuration.Project instance, "
                f"not {type(project).__name__}"
            )
        self.reports = {}
        for step in [
            "configure",
            "build",
            "install",
            "clean",
            "test",
            "script",
        ]:
            self.reports[step] = Report(f"{__name__}_{step}.cmake_new")
        self.build_logs = {}

    @property
    def build_dir(self):
        return Path(self.project.name) / "build"

    def _cache_preload_file(self) -> Path:
        """
        Name of the cache preload file to be passed to CMake.
        """
        return self.build_dir / "cache_preload.cmake"

    def _prepare_cache(
        self,
        cache_entries=None,
    ):
        """
        Prepare the cache_preload.cmake file passed to CMake during the
        configuration.
        """
        # prepare the cache to give to CMake: add the launcher rules commands,
        # followed by what is found passed as argument
        if cache_entries is None:
            cache_entries = []
        elif hasattr(cache_entries, "items"):
            cache_entries = list(cache_entries.items())

        cache_file = self._cache_preload_file()
        ensure_dir(os.path.dirname(cache_file), self.reports["configure"])
        with open(cache_file, "w") as cache:
            cache.writelines(
                [
                    'set(%s "%s" CACHE STRING "override")\n' % item
                    for item in cache_entries
                ]
            )
            # force use of ccache
            cache.writelines(
                'set(CMAKE_{}_COMPILER_LAUNCHER ccache CACHE STRING "override")\n'.format(
                    lang
                )
                for lang in ("C", "CXX")
            )
            # enable rule wrappers
            cache.writelines(
                (
                    'set(CMAKE_RULE_LAUNCH_{} "lb-wrapcmd <CMAKE_CURRENT_BINARY_DIR> <TARGET_NAME>"\n'
                    '    CACHE STRING "override")\n'
                ).format(action)
                for action in ("COMPILE", "LINK", "CUSTOM")
            )

    def _env(self, init_env=None, override_env=None):
        """
        Compute the environment for the project.
        """
        new_env = self.project.environment(envdict=init_env)
        if override_env:
            new_env.update(override_env)
        # make sure that slot deployment directory is the first one in CMAKE_PREFIX_PATH
        if self.project.slot:
            cmake_paths = new_env.get("CMAKE_PREFIX_PATH", "").split(":")
            deployment_dir = str(self.project.slot.get_deployment_directory())
            if cmake_paths[0] != deployment_dir:
                new_env["CMAKE_PREFIX_PATH"] = ":".join(
                    [deployment_dir]
                    + [item for item in cmake_paths if item != deployment_dir]
                )
        return new_env

    def _run(self, cmd, **kwargs):
        # strip special kwargs before command invocation
        cmd_kwargs = {n: v for n, v in kwargs.items() if n not in self.SPECIAL_ARGS}
        cmd_kwargs["env"] = self._env(
            init_env=kwargs.get("init_env"), override_env=kwargs.get("override_env")
        )
        cmd_kwargs["cwd"] = str(kwargs.get("cwd", Path.cwd()))
        cmd = [str(c) for c in cmd]
        started = datetime.now()
        result = singularity_run(
            cmd,
            cmd_kwargs["env"],
            cmd_kwargs["cwd"],
            task={
                "task": kwargs.get("step") if kwargs.get("step") == "test" else "build",
                "project": self.project.id(),
                "platform": cmd_kwargs["env"]["BINARY_TAG"],
                "worker_task_id": kwargs.get("worker_task_id"),
            },
            krb_auth=kwargs.get("krb_auth", None),
        )
        if kwargs.get("step") == "build":
            self.build_logs.update(result.build_logs)

        completed = datetime.now()
        step = kwargs.get("step", "script")
        self.reports[step].info(f"running {' '.join(cmd)}")
        self.reports[step].info(f"command exited with code {result.returncode}")
        self.reports[step].started = started.isoformat()
        self.reports[step].completed = completed.isoformat()
        self.reports[step].returncode = result.returncode
        self.reports[step].command = " ".join(quote(a) for a in cmd)
        self.reports[step].stderr = result.stderr
        self.reports[step].stdout = (
            (
                f"#### {self} {step} ####\n"
                f"# Start: {started.isoformat()}\n"
                f"# Command: {' '.join(quote(a) for a in cmd)}\n"
            )
            + result.stdout.decode(errors="surrogateescape")
            + (
                f"# Return code: {result.returncode}\n"
                f"# End: {completed.isoformat()}\n"
            )
        ).encode(errors="surrogateescape")

    def configure(self, **kwargs):
        self._prepare_cache(
            cache_entries=kwargs.get("cache_entries"),
        )
        cmd = [
            "cmake",
            "-S",
            self.project.get_deployment_directory(),
            "-B",
            self.build_dir,
            "-G",
            kwargs.get("generator", "Ninja"),
            "-C",
            self._cache_preload_file(),
        ]
        # get the toolchain to use
        if self.project.slot and hasattr(self.project.slot, "LCG"):
            LCG_VERSION = self.project.slot.LCG.version
        elif "LCG_VERSION" in kwargs.get("init_env", {}):
            LCG_VERSION = kwargs["init_env"]["LCG_VERSION"]
        else:
            out = (
                "\033[0;31mslot configuration error: "
                "version of LCG not defined (required for new "
                "CMake configuration)\033[0m\n"
            )
            raise EnvironmentError(out)

        # When the configuration sets LCG_EXTERNALS_FILE, we have to use the
        # toolchain special/lcg-nightly.cmake
        if "LCG_EXTERNALS_FILE" not in kwargs.get("cache_entries", []):
            toolchain = find_path(
                os.path.join(
                    "lcg-toolchains",
                    "LCG_{}".format(LCG_VERSION),
                    "{}.cmake".format(kwargs["init_env"]["BINARY_TAG"]),
                ),
                search_path=self._env(kwargs.get("init_env"))
                .get("CMAKE_PREFIX_PATH", "")
                .split(os.pathsep),
            )
        else:
            toolchain = find_path(
                os.path.join("lcg-toolchains", "special", "lcg-nightly.cmake"),
                search_path=self._env(kwargs.get("init_env"))
                .get("CMAKE_PREFIX_PATH", "")
                .split(os.pathsep),
            )

        if not toolchain:
            out = (
                "\033[0;31mslot configuration error: "
                "cannot find toolchain file for {} {}\033[0m\n"
            ).format(LCG_VERSION, kwargs["init_env"]["BINARY_TAG"])
            raise FileNotFoundError(out)

        cmd.append("-DCMAKE_TOOLCHAIN_FILE=" + toolchain)

        self._run(cmd, step="configure", **kwargs)
        self.build_logs.setdefault("configure", {}).setdefault("configure", []).append(
            {
                "cmd": [str(c) for c in cmd],
                "returncode": self.reports["configure"].returncode,
                "stdout": self.reports["configure"].stdout.decode(
                    errors="surrogateescape"
                ),
                "stderr": self.reports["configure"].stderr.decode(
                    errors="surrogateescape"
                ),
                "started": self.reports["configure"].started,
                "completed": self.reports["configure"].completed,
            }
        )

    def install(self, **kwargs):
        return self._run(
            [
                "cmake",
                "--install",
                self.build_dir,
                "--prefix",
                Path(self.project.name)
                / "InstallArea"
                / kwargs["init_env"]["BINARY_TAG"],
            ],
            step="install",
            **kwargs,
        )

    def build(self, **kwargs):
        """
        Build a project in the current directory using new CMake configuration.

        project has to be a Project instance from lb.nightly.configuration

        Return logging data in the form of a list of messages and executed commands.
        """
        self.configure(**kwargs)
        if self.reports["configure"].returncode:
            # no point trying to build if we failed to configure
            return self.reports, self.build_logs

        cmd = ["cmake", "--build", self.build_dir]
        if "jobs" in kwargs:
            cmd.extend(["-j", str(kwargs["jobs"])])
        if "args" in kwargs:
            cmd.append("--")
            cmd.extend((arg if arg != "-k" else "-k0") for arg in kwargs["args"])
        self._run(cmd, step="build", **kwargs)
        self.install(**kwargs)
        return self.reports, self.build_logs

    def clean(self, **kwargs):
        return self._run(
            [
                "cmake",
                "--build",
                self.build_dir,
                "--target",
                "clean",
            ],
            step="clean",
            **kwargs,
        )

    def test(self, **kwargs):
        cmd = ["ctest", "-T", "test"]
        if "jobs" in kwargs:
            cmd.extend(["-j", str(kwargs["jobs"])])
        self._run(
            cmd,
            step="test",
            cwd=self.build_dir,
            **kwargs,
        )
        return self.reports

    def __str__(self):
        return "CMake (new)"


def is_new_cmake_style(project_dir: Path):
    """
    Check if the project uses the old or new CMake configuration style.
    """
    top_config = project_dir / "CMakeLists.txt"
    if not top_config.exists():
        # no CMakeLists.txt -> it's a CMT project
        return False
    if (project_dir / "toolchain.cmake").exists():
        # custom toolchain file in the sources -> must be old style
        return False
    with open(top_config) as f:
        content = f.read()
    # new style projects do not call "find_package(GaudiProject)"
    return not bool(re.search(r"find_package\s*\(\s*GaudiProject", content))


class make:
    """
    Base class for build tools based on make.
    """

    def __init__(self, project: Project) -> None:
        self.project = project
        if not isinstance(self.project, Project):
            raise TypeError(
                f"expected lb.nightly.configuration.Project instance, "
                f"not {type(project).__name__}"
            )
        self.build_logs = {}

    def _make(self, target, **kwargs):
        """
        Internal function to wrap the call to make for CMT.

        @param target: name of the target to build
        @param jobs: number of parallel build processes [default: 1]
        @param env: dictionary used to override environment variables from the
                    project configuration
        @param args: list of extra arguments to pass to make
        @param make_cmd: command to be used to build [default: ['make']]
        """
        report = Report(f"{__name__}.make")
        jobs = kwargs.get("jobs")

        env = self.project.environment(envdict=kwargs.get("init_env"))
        env.update(kwargs.get("override_env", {}))
        # "unset" variables set to None
        env = dict((key, value) for key, value in env.items() if value is not None)

        cmd_kwargs = {"env": env, "cwd": self.project.baseDir}
        if "stderr" in kwargs:
            cmd_kwargs["stderr"] = kwargs["stderr"]

        cmd = kwargs.get("make_cmd") or "make"
        if isinstance(cmd, str):
            cmd = cmd.split()
        else:
            # make a copy of make_cmd argumens to avoid modifying it
            cmd = list(cmd)

        if jobs:
            cmd.append("-j%d" % jobs)
        cmd.extend(kwargs.get("args", []))
        cmd.append(target)

        ensure_dir(cmd_kwargs["cwd"], report)

        init_project(cmd_kwargs["cwd"])
        report.info("running %s", " ".join(cmd))
        started = datetime.now()
        result = singularity_run(
            cmd,
            cmd_kwargs["env"],
            cmd_kwargs["cwd"],
            task={
                "task": "build",
                "project": self.project.id(),
                "platform": cmd_kwargs["env"].get("BINARY_TAG"),
                "worker_task_id": kwargs.get("worker_task_id"),
            },
            krb_auth=kwargs.get("krb_auth", None),
        )
        if kwargs.get("step") == "all":
            self.build_logs.update(result.build_logs)

        completed = datetime.now()
        report.info("command exited with code %d", result.returncode)

        report.stdout = (
            f"#### {self} {target} ####\n"
            f"# Start: {started.isoformat()}\n"
            f"# Command: {' '.join(quote(a) for a in cmd)}\n"
            + result.stdout.decode(errors="surrogateescape")
            + f"# End: {completed.isoformat()}\n"
        ).encode(errors="surrogateescape")
        report.stderr = result.stderr
        report.returncode = result.returncode
        report.started = started.isoformat()
        report.completed = completed.isoformat()
        report.command = " ".join(quote(a) for a in cmd)
        return report

    def build(self, **kwargs):
        """
        Build a project.
        """
        return self._make("all", **kwargs)

    def clean(self, **kwargs):
        """
        Clean the build products.
        """
        return self._make("clean", **kwargs)

    def test(self, **kwargs):
        """
        Run the tests.
        """
        return self._make("test", **kwargs)

    def __str__(self):
        """
        Conversion to string.
        """
        return "make"


class cmake_old(make):
    def __init__(self, project: Project) -> None:
        self.project = project
        if not isinstance(self.project, Project):
            raise TypeError(
                f"expected lb.nightly.configuration.Project instance, "
                f"not {type(project).__name__}"
            )
        self.reports = {}
        for step in [
            "configure",
            "build",
            "install",
            "clean",
            "test",
            "script",
        ]:
            self.reports[step] = Report(f"{__name__}_{step}.cmake_old")
        self.build_logs = {}

    def _cache_preload_file(self):
        """
        Name of the cache preload file to be passed to CMake.
        """
        return os.path.join(self.project.baseDir, "cache_preload.cmake")

    def _prepare_cache(self, cache_entries=None):
        """
        Prepare the cache_preload.cmake file passed to CMake during the
        configuration.
        """
        # prepare the cache to give to CMake: add the launcher rules commands,
        # followed by what is found passed as argument
        if cache_entries is None:
            cache_entries = []
        elif hasattr(cache_entries, "items"):
            cache_entries = list(cache_entries.items())

        # add the RULE_LAUNCH settings for the build
        launcher_cmd = "lb-wrapcmd <CMAKE_CURRENT_BINARY_DIR> <TARGET_NAME>"
        cache_entries = [
            ("GAUDI_RULE_LAUNCH_%s" % n, launcher_cmd)
            for n in ("COMPILE", "LINK", "CUSTOM")
        ] + cache_entries

        cache_file = self._cache_preload_file()
        ensure_dir(os.path.dirname(cache_file), self.reports["configure"])
        with open(cache_file, "w") as cache:
            cache.writelines(
                [
                    'set(%s "%s" CACHE STRING "override")\n' % item
                    for item in cache_entries
                ]
            )

    def _make(self, target, **kwargs):
        """
        Override basic make call to set the environment variable USE_CMT=1.
        """
        # copy kwargs to be able to change it
        kwargs = dict(kwargs)
        self._prepare_cache(cache_entries=kwargs.pop("cache_entries", None))

        env = kwargs.pop("init_env", {})
        preload_file = os.path.join(os.getcwd(), self._cache_preload_file())
        env.update(
            {
                "USE_CMAKE": "1",
                "USE_MAKE": "1",
            }
        )
        try:
            kwargs["make_cmd"] = kwargs["make_cmd"].get(target)
        except (KeyError, TypeError, AttributeError):
            # no target-specific make_cmd
            pass
        return make._make(self, target, env=env, **kwargs)

    def build(self, **kwargs):
        """
        Override the basic build method to call the different targets used in
        CMake builds: configure, all, unsafe-install, post-install.
        """
        kwargs.setdefault("args", [])
        kwargs["args"].insert(0, "BUILDDIR=build")
        for target in (
            "configure",
            "all",
            "unsafe-install",
            "post-install",
            "clean",
        ):
            self.reports[target] = self._make(target, **kwargs)

        # Write configuration logfile fragment for "collect logs" script
        ensure_dir(
            os.path.join(self.project.baseDir, "build", "configure"),
            self.reports["configure"],
        )
        compl = datetime.fromisoformat(self.reports["configure"].completed)
        start = datetime.fromisoformat(self.reports["configure"].started)
        with open(
            os.path.join(
                self.project.baseDir,
                "build",
                "configure",
                "{:%s%f000}-build.log".format(compl),
            ),
            "w",
        ) as conf_log:
            conf_log.writelines(self.reports["configure"].stdout.splitlines(True)[2:-1])
            conf_log.write(
                "\033[0;34mConfiguration completed in {} seconds\033[0m\n".format(
                    (compl - start).total_seconds()
                )
            )
        self.build_logs.setdefault("configure", {}).setdefault("configure", []).append(
            {
                "cmd": (self.reports["configure"].command).split(),
                "returncode": self.reports["configure"].returncode,
                "stdout": "\n".join(
                    self.reports["configure"]
                    .stdout.decode(errors="surrogateescape")
                    .splitlines(True)[2:-1]
                ),
                "stderr": self.reports["configure"].stderr.decode(
                    errors="surrogateescape"
                ),
                "started": self.reports["configure"].started,
                "completed": self.reports["configure"].completed,
            }
        )

        return self.reports, self.build_logs

    def clean(self, **kwargs):
        """
        Override default clean method to call the 'purge' target (more
        aggressive).
        """
        return self._make("purge", **kwargs)

    def test(self, **kwargs):
        """
        Run the tests in a Gaudi/LHCb project.
        """
        kwargs.setdefault("args", [])
        kwargs["args"].insert(0, "BUILDDIR=build")
        if "jobs" in kwargs:
            flag = "-j{}".format(kwargs.pop("jobs"))
            for i, arg in enumerate(kwargs["args"]):
                if arg.startswith("ARGS="):
                    kwargs["args"][i] += " " + flag
                    break
            else:  # (this else matches the for) no ARGS= in args
                kwargs["args"].append(f"ARGS={flag}")
        for target in ("configure", "test"):
            self.reports[target] = self._make(target, **kwargs)
        return self.reports

    def __str__(self):
        """
        Conversion to string.
        """
        return "CMake (old)"


class cmt(make):
    """
    Class to wrap the build/test semantics for CMT-based projects.
    """

    def _make(self, target, **kwargs):
        """
        Override basic make call to set the environment variable USE_CMT=1.
        """
        env = kwargs.pop("init_env", {})
        # PWD and CWD may cause troubles to CMT, so we unset them
        env.update({"USE_CMT": "1", "PWD": None, "CWD": None})
        if "CMTROOT" not in env and "make_cmd" not in kwargs:
            kwargs["make_cmd"] = ["cmt", "run", "make"]
        return make._make(self, target, env=env, **kwargs)

    def clean(self, **kwargs):
        """
        Override default clean method to call the 'purge' target (more
        aggressive).
        """
        return self._make("purge", **kwargs)

    def test(self, **kwargs):
        """
        Run the tests in a Gaudi/LHCb project using CMT.
        """
        env = kwargs.get("init_env", {})
        if "CMTCONFIG" not in env:
            env["CMTCONFIG"] = env.get("BINARY_TAG")
        if "GAUDI_QMTEST_HTML_OUTPUT" not in env:
            bin_dir = os.path.join(
                os.path.abspath(self.project.baseDir), "build", "html"
            )
            env["GAUDI_QMTEST_HTML_OUTPUT"] = bin_dir
        kwargs["env"] = env
        return self._make("test", **kwargs)

    def __str__(self):
        """
        Conversion to string.
        """
        return "CMT"


class cmake:
    """
    Dispatcher to use the old or new style CMake build procedure
    depending on the project.
    """

    def __new__(cls, project: Project):
        return (
            cmake_new(project)
            if is_new_cmake_style(project.get_deployment_directory())
            else cmake_old(project)
        )

    def __str__(self):
        """
        Conversion to string.
        """
        return "CMake"



###############################################################################
# (c) Copyright 2020-2021 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import os
from typing import Union

from git import Git, GitCommandError, Repo
from lb.nightly.configuration import DataProject, Package, Project

from .common import Report
from .gitlab_helpers import get_gitlab_id, get_gitlab_url


def git(project: Union[Project, Package]):
    """
    Checkout a project in the current directory.

    project has to be a Project instance from lb.nightly.configuration

    Return logging data in the form of a list of messages and executed commands.
    """
    if not isinstance(project, (Project, Package)):
        raise TypeError(
            f"expected lb.nightly.configuration.Project instance, not {type(project).__name__}"
        )

    rep = Report(f"{__name__}.git")
    rep.project = {
        "name": project.name,
        "version": project.version,
        "options": project.checkout_opts,
    }
    rep.merges = {"success": [], "failure": []}
    rep.submodules = {"success": [], "failure": []}
    rep.commit = None
    rep.tree = None

    if isinstance(project, DataProject):
        # Data projects must be handled in a special way
        rep.packages = [git(pack) for pack in project.packages]
        return rep

    url = get_gitlab_url(project)
    rep.gitlab_id = get_gitlab_id(url)
    if hasattr(project, "container") and project.container:
        path = os.path.join(
            os.getcwd(), project.container.name, project.name, project.version
        )
    else:
        path = os.path.join(os.getcwd(), project.name)

    def submodule_recursion(submodules, path=path):
        """
        Call "git update" on the passed list of submodules
        and all their children.
        """
        for s in submodules:
            target_path = os.path.join(path, s.path)
            rep.info(f"updating submodule {s.name} in {target_path}")
            try:
                s.update()
                rep.submodules["success"].append(s.name)
                submodule_recursion(s.children(), target_path)
            except GitCommandError as err:
                rep.submodules["failure"].append(s.name)
                rep.git_error(f"{s.name}: warning: failed to get submodule", err)

    rep.info(f"cloning {url} into {path}")

    if not os.path.exists(path):
        repo = Repo.clone_from(url, path, no_checkout=True)
    else:
        repo = Repo(path)

    # pick up requested commit
    if "commit" in project.checkout_opts:
        repo.head.reference = repo.commit(project.checkout_opts["commit"])
    elif project.version.lower() == "head":
        pass  # trust default commit
    elif project.version in repo.remotes.origin.refs:  # check if version is a branch
        repo.head.reference = repo.remotes.origin.refs[project.version]
    elif project.version in repo.refs:  # check if version is a tag
        repo.head.reference = repo.refs[project.version]
    else:
        raise ValueError(f"no 'commit' option and invalid version for {project}")
    rep.info(f"using commit {repo.head.commit.hexsha} for {project.version}")

    repo.head.reset(index=True, working_tree=True)
    git_cmd = Git(repo.working_dir)

    for mr_iid, commit_id in project.checkout_opts.get("merges", []):
        try:
            rep.info(f"merging {rep.gitlab_id}!{mr_iid} ({commit_id})")
            fetches = repo.remote("origin").fetch(commit_id)
            commit = fetches[0].commit
            rep.log(
                "stdout",
                "debug",
                git_cmd.merge(
                    commit, no_ff=True, message=f"merged {rep.gitlab_id}!{mr_iid}"
                ),
            )
            rep.merges["success"].append(mr_iid)
        except GitCommandError as err:
            rep.git_error(
                f"warning: failed to merge {rep.gitlab_id}!{mr_iid} ({commit_id})", err
            )
            rep.info("reverting to previous state")
            repo.head.reset(index=True, working_tree=True)
            rep.merges["failure"].append(mr_iid)

    submodule_recursion(repo.submodules)

    rep.commit = repo.head.commit.hexsha
    rep.tree = repo.head.commit.tree.hexsha
    rep.info("result:\n  commit: %s\n  tree: %s", rep.commit, rep.tree)

    return rep


def notify_gitlab(project: Union[Project, Package], report: Report, token: str = None):
    """
    Update merge requests discussions with links to the slot the project
    was built in, specifying if the merge was successful or not.

    The ``report`` argument should be the return value of a call to
    the ``git`` function.
    """
    from .gitlab_helpers import notifyMergeRequest

    if not hasattr(report, "gitlab_id"):
        report.gitlab_id = get_gitlab_id(get_gitlab_url(project))

    report.debug("notifying gitlab project %s", report.gitlab_id)
    try:
        for mr_iid in report.merges.get("success", []):
            notifyMergeRequest(project, report.gitlab_id, mr_iid, True, token, report)
        for mr_iid in report.merges.get("failure", []):
            notifyMergeRequest(project, report.gitlab_id, mr_iid, False, token, report)
    except AttributeError:
        report.warning("summary of merge results missing in report")

    return report


def update_nightly_git_archive(project: Union[Project, Package], report: Report):
    """
    Record the archives of the nightly builds in
    https://gitlab.cern.ch/lhcb-nightlies

    A new tags from slot name and build id is created, and the slot
    branch is updated, if needed.

    The ``report`` argument should be the return value of a call to
    the ``git`` function.
    """
    from .gitlab_helpers import gitlabProjectExists

    assert (
        project.slot and project.slot.build_id
    ), "Only projects in built slots can be recorded in https://gitlab.cern.ch/lhcb-nightlies"

    if not gitlabProjectExists(f"lhcb-nightlies/{project.name}"):
        report.warning(
            f"{project.name} not found on https://gitlab.cern.ch/lhcb-nightlies, not recording build sources"
        )
        return report

    report.info(
        f"recording checkout state in https://gitlab.cern.ch/lhcb-nightlies/{project.name}"
    )
    repo = Repo(project.name)
    nightlies = repo.create_remote(
        "nightlies", f"ssh://git@gitlab.cern.ch:7999/lhcb-nightlies/{project.name}.git"
    )

    # fetch guessed "previous" tag (build_id - 1)
    try:
        previous_tag = repo.create_tag(
            f"{project.slot.name}/{project.slot.build_id - 1}",
            nightlies.fetch(f"{project.slot.name}/{project.slot.build_id - 1}")[
                0
            ].commit,
        )
    except GitCommandError:
        previous_tag = None

    # fetch the slot specific branch from the archive
    try:
        tracking_branch = repo.create_head(
            project.slot.name, nightlies.fetch(project.slot.name)[0].commit
        )
    except GitCommandError:
        report.debug(
            f"failed to fetch branch {project.slot.name} from {nightlies.url}, trying to create it"
        )
        if previous_tag:
            tracking_branch = repo.create_head(project.slot.name, previous_tag.commit)
        else:
            tracking_branch = repo.create_head(project.slot.name, repo.head.commit)

    if previous_tag and previous_tag.commit.tree.hexsha == report.tree:
        # same content of previous tag, just re-use that commit
        repo.head.reset(previous_tag.commit)
    elif tracking_branch.commit.tree.hexsha == report.tree:
        # same content as tip of reference branch, re-use that commit
        repo.head.reset(tracking_branch.commit)
    else:
        # no commit to re-use, create a new merge commit on top of the
        repo.index.commit(
            f"changes for {project.slot.name}/{project.slot.build_id}",
            parent_commits=(tracking_branch.commit, repo.head.commit),
        )
        tracking_branch.commit = repo.head.commit
    # create the new tag
    new_tag = repo.create_tag(f"{project.slot.name}/{project.slot.build_id}")

    assert (
        report.tree == repo.head.commit.tree.hexsha
    ), "tree content mismatch after nightlies archive update"

    nightlies.push([tracking_branch, new_tag])

    # log changes, if any
    if previous_tag and previous_tag.commit.tree.hexsha != report.tree:
        report.info(f"changes detected wrt {previous_tag.name}")
        report.log(
            "command",
            "debug",
            f"git show-branch --sha1-name {previous_tag.name} {new_tag.name}",
        )
        report.log(
            "stdout",
            "debug",
            repo.git.show_branch([previous_tag, new_tag], sha1_name=True, color=True),
        )
        report.log(
            "command",
            "debug",
            f"git diff --stat {previous_tag.name} {new_tag.name}",
        )
        # find a reasonable report width to avoid truncation of names
        # - use the biggest between 120 and the max size of names + 20
        width = max(
            120,
            max(
                len(l)
                for l in repo.git.diff(
                    [previous_tag, new_tag], z=True, name_only=True
                ).split("\0")
            )
            + 20,
        )
        report.log(
            "stdout",
            "debug",
            repo.git.diff(
                [previous_tag, new_tag],
                stat=f"{max(width + 20, 120)}",
                color=True,
            ),
        )
    elif previous_tag:
        report.info(f"no change wrt {previous_tag.name}")
    else:
        report.info(f"no previous tag found, comparison not available")

    # update the commit id in the report (the tree did not change by
    # construction)
    report.commit = repo.head.commit.hexsha

    return report



###############################################################################
# (c) Copyright 2020-2021 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import html
import os
import re

MAX_REPORTED_ISSUES = 100  # per severity
_ESCAPE_SEQ = re.compile("\x1b\\[([0-9;]*m|[012]?K)")


def remove_colors(text):
    """
    Strip ANSI color codes from a text.

    >>> remove_colors('\\x1b[34;42mHello\\x1b[2m!\\x1b[m')
    'Hello!'
    """
    return _ESCAPE_SEQ.sub("", text)


_LOG_SCANNERS = []


def log_scanner(func):
    """
    Decorator to declare a function as log scanner.
    """
    global _LOG_SCANNERS
    _LOG_SCANNERS.append(func)
    return func


class IssueSource(object):
    def __init__(self, name, line=None, pos=None):
        self.name = name
        self.line = line
        self.pos = pos

    def __str__(self):
        return ":".join(str(s) for s in [self.name, self.line, self.pos] if s)

    def __repr__(self):
        return "IssueSource{0!r}".format(
            tuple(s for s in [self.name, self.line, self.pos] if s)
        )


class Issue:
    """
    Base class for issues found in logs.
    """

    SEVERITIES = ("error", "warning", "coverity")

    def __init__(self, severity, source, msg, log_range):
        if severity not in self.SEVERITIES:
            raise ValueError("invalid severity value %r", severity)
        self.severity = severity
        if isinstance(source, tuple):
            source = IssueSource(*source)
        self.source = source
        self.msg = msg
        self.log_range = log_range

    def linkText(self):
        return "{}:{}".format(self.source.name, self.source.line)

    def __str__(self):
        return ": ".join(str(s) for s in [self.source, self.severity, self.msg])

    def __repr__(self):
        return "{0}{1!r}".format(
            self.__class__.__name__,
            (self.severity, self.source, self.msg, self.log_range),
        )


class IssueLinker:
    def __init__(self, issues, url_format):
        self.issues = issues
        self.url_format = url_format
        self._count = 0

    def __call__(self, line):
        n = self._count
        for issue in self.issues:
            l, h = issue.log_range
            if n >= l and n < h:
                txt = html.escape(issue.linkText(), quote=True)
                line = line.replace(
                    txt,
                    '<a href="{url}">{txt}</a>'.format(
                        url=self.url_format(issue.source), txt=txt
                    ),
                )
                break
        self._count += 1
        return line


class GitlabUrlFormat:
    def __init__(self, base_url):
        base_url = base_url.rstrip("/")
        self.format = base_url + "/{name}#L{line}"

    def __call__(self, source):
        return self.format.format(name=source.name, line=source.line)


def reports2exclusions(reports=None):
    """
    Translate a list of Issue instances to a list of lines to exclude in an
    enumeration.

    >>> reports2exclusions([Issue('warning', ('f1',), 'm1', (10, 15)),
    ...                     Issue('warning', ('f2',), 'm2', (23, 27)),
    ...                     Issue('error', ('f3',), 'm3', (18, 19))])
    [(10, 15), (18, 19), (23, 27)]
    """
    if not reports:
        return None
    exclusions = [issue.log_range for issue in reports]
    exclusions.sort()
    return exclusions


def enumerate_x(items, exclusions=None):
    """
    Same as builtin enumerate function, but skip items with id in the ranges
    defined by exclusions.

    >>> list(enumerate_x('abcdefghij', [(2, 5), (8, 9)]))
    [(0, 'a'), (1, 'b'), (5, 'f'), (6, 'g'), (7, 'h'), (9, 'j')]

    Exclusions may be overlapping:

    >>> list(enumerate_x('abcdefghij', [(1, 4), (2, 5), (6, 9), (7, 8)]))
    [(0, 'a'), (5, 'f'), (9, 'j')]
    """
    if not exclusions:
        return enumerate(items)
    from itertools import chain

    # invert the exclusions to list of partial enumerations
    ranges = []
    start = 0
    for stop, next_start in exclusions:
        # if there is an overlap in the exclusion we need to skip to the next
        if stop >= start:
            ranges.append(enumerate(items[start:stop], start))
        # max here covers the case of full enclosed exclusions ([(2,10), (3,5)])
        start = max(start, next_start)
    ranges.append(enumerate(items[start:], start))
    # return the joined partial enumerations
    return chain(*ranges)


class GCCIssue(Issue):
    pass


@log_scanner
def gcc_output_scanner(lines, reports=None):
    if reports is None:
        reports = []
    diag_rexp = re.compile(r"^(\S+):([0-9]+):([0-9]+): (warning|error): (.*)")
    try:
        ln_iter = enumerate_x(lines, reports2exclusions(reports))
        ln, line = next(ln_iter)
        while True:
            line = remove_colors(line.rstrip())
            m = diag_rexp.match(line)
            if m:
                startln = ln
                ln, line = next(ln_iter)
                try:
                    while line.startswith(" "):
                        ln, line = next(ln_iter)
                except StopIteration:
                    pass
                reports.append(
                    GCCIssue(
                        m.group(4),
                        (m.group(1), int(m.group(2)), int(m.group(3))),
                        m.group(5),
                        (startln, ln),
                    )
                )
            else:
                ln, line = next(ln_iter)
    except StopIteration:
        pass
    return reports


class CMakeIssue(Issue):
    pass


@log_scanner
def cmake_output_scanner(lines, reports=None):
    if reports is None:
        reports = []
    diag_rexp = re.compile(
        r"^CMake (?:Deprecation )?(Warning|Error) (?:\(dev\) )?at (\S+):([0-9]+) \(message\):"
    )
    try:
        ln_iter = enumerate_x(lines, reports2exclusions(reports))
        ln, line = next(ln_iter)
        while True:
            line = line.rstrip()
            m = diag_rexp.match(line)
            if m:
                startln = ln
                ln, line = next(ln_iter)
                try:
                    while line.startswith(" "):
                        ln, line = next(ln_iter)
                except StopIteration:
                    pass
                msg = " ".join(l.strip() for l in lines[startln + 1 : ln]).strip()
                if len(msg) > 120:
                    msg = msg[:80] + "[...]" + msg[-35:]
                reports.append(
                    CMakeIssue(
                        m.group(1).lower(),
                        (m.group(2), int(m.group(3))),
                        msg,
                        (startln, ln),
                    )
                )
            else:
                ln, line = next(ln_iter)
    except StopIteration:
        pass
    return reports


class GaudiIssue(Issue):
    pass


@log_scanner
def gaudi_output_scanner(lines, reports=None):
    if reports is None:
        reports = []
    diag_rexp = re.compile(r".*?(\S+) *(WARNING|ERROR|FATAL) +(.*)")
    try:
        ln_iter = enumerate_x(lines, reports2exclusions(reports))
        ln, line = next(ln_iter)
        while True:
            line = line.rstrip()
            m = diag_rexp.match(line)
            if m:
                diag_type = m.group(2).lower()
                if diag_type == "fatal":
                    diag_type = "error"
                reports.append(
                    GaudiIssue(diag_type, (m.group(1),), m.group(3), (ln, ln + 1))
                )
            ln, line = next(ln_iter)
    except StopIteration:
        pass
    return reports


class PythonIssue(Issue):
    def linkText(self):
        return 'File "{}", line {}'.format(self.source.name, self.source.line)


@log_scanner
def python_output_scanner(lines, reports=None):
    if reports is None:
        reports = []
    warn_rexp = re.compile(r"(\S+):([0-9]+): \S*(Warning): (.*)")
    tbinfo_rexp = re.compile(r'\s+File "(.+)", line ([0-9]+)')
    try:
        ln_iter = enumerate_x(lines, reports2exclusions(reports))
        ln, line = next(ln_iter)
        while True:
            line = line.rstrip()
            m = warn_rexp.match(line)
            if m:
                startln = ln
                ln, line = next(ln_iter)
                reports.append(
                    PythonIssue(
                        "warning",
                        (m.group(1), int(m.group(2))),
                        m.group(4),
                        (startln, ln),
                    )
                )
            elif line.strip() == "Traceback (most recent call last):":
                startln = ln
                ln, line = next(ln_iter)
                filename, fileln, msg = "<unkown>", None, "Python traceback"
                try:
                    while line.startswith(" "):
                        ln, line = next(ln_iter)
                        m = tbinfo_rexp.match(line)
                        if m:
                            filename = m.group(1)
                            fileln = int(m.group(2))
                    # the actual exception message is after of the dump
                    msg = line.strip()
                    ln, line = next(ln_iter)
                except StopIteration:
                    pass
                reports.append(
                    PythonIssue("error", (filename, fileln), msg, (startln, ln))
                )
            else:
                ln, line = next(ln_iter)
    except StopIteration:
        pass
    return reports


class ROOTBreakIssue(Issue):
    pass


@log_scanner
def root_stack_trace_output_scanner(lines, reports=None):
    if reports is None:
        reports = []
    MARKER = "*** Break ***"
    try:
        ln_iter = enumerate_x(lines, reports2exclusions(reports))
        ln, line = next(ln_iter)
        while True:
            line = line.strip()
            if line.startswith(MARKER):
                msg = line[len(MARKER) :].strip()
                reports.append(
                    ROOTBreakIssue("error", ("<unknown>",), msg, (ln, ln + 1))
                )
            ln, line = next(ln_iter)
    except StopIteration:
        pass
    return reports


class GenericIssue(Issue):
    def __str__(self):
        return self.msg


@log_scanner
def generic_scanner(lines, reports=None):
    if reports is None:
        reports = []
    WARNING_RE = re.compile(
        "|".join([r"\bwarning\b", r"\bSyntaxWarning:"]), re.IGNORECASE
    )
    ERROR_RE = re.compile(
        "|".join(
            [
                r"\berror\b",
                r"\*\*\* Break \*\*\*",
                r"^Traceback \(most recent call last\):",
                r"^make: \*\*\* No rule to make target",
                r"Assertion.*failed",
            ]
        ),
        re.IGNORECASE,
    )
    try:
        ln_iter = enumerate_x(lines, reports2exclusions(reports))
        ln, line = next(ln_iter)
        while True:
            line = remove_colors(line.strip())
            diag_type = (WARNING_RE.search(line) and "warning") or (
                ERROR_RE.search(line) and "error"
            )
            if diag_type:
                reports.append(
                    GenericIssue(diag_type, ("<unknown>",), line, (ln, ln + 1))
                )
            ln, line = next(ln_iter)
    except StopIteration:
        pass
    return reports


def scan_build_log(lines):
    reports = None
    for scanner in _LOG_SCANNERS:
        reports = scanner(lines, reports)
    return reports


def report_dict(chunks, reports):
    """
    Convert list of (chunk_id, lines) and a dictionary {chunk_id: issues} into
    a simplified report suitable for conversion to JSON.

    The output format is:

        {'sections': [{'id': 'section_id',
                       'desc': 'html title of the section',
                       'url': 'url of section content (html)'}, ...],
         'issues': {
             'severity': [{
                  'section_id': 'id of section it comes from',
                  'anchor': 'html anchor in section content',
                  'desc': 'one line description of the issue',
                  'text': ['full issue', 'report', ...]
             }, ...],
             ...
        }}
    """
    report = {"sections": [], "issues": {severity: [] for severity in Issue.SEVERITIES}}

    for chunk_id, chunk in chunks:
        report["sections"].append(
            {"id": chunk_id, "desc": chunk_id, "url": "{}.html".format(chunk_id)}
        )
        for issue in reports[chunk_id]:
            report["issues"][issue.severity].append(
                {
                    "section_id": chunk_id,
                    "anchor": "{0}_{1}".format(chunk_id, issue.log_range[0] + 1),
                    "desc": html.escape(str(issue), quote=True),
                    "text": [
                        html.escape(remove_colors(line.rstrip()), quote=True)
                        for line in chunk[issue.log_range[0] : issue.log_range[1]]
                    ],
                }
            )
    return report


TABLE_HEAD = '<table class="table table-striped" style="text-align:left">'
TABLE_TAIL = "</table>"


def generate_build_report(
    build_logs={},
    proj_build_root=None,
    exceptions=None,
    extra_info=None,
):
    from os.path import isdir, join
    from subprocess import CalledProcessError, check_output

    # collect git infos
    git_info = {}
    try:
        if proj_build_root and isdir(join(proj_build_root, ".git")):
            # we can only support projects with checkout recorded in
            # https://gitlab.cern.ch/lhcb-nightlies
            url = check_output(
                ["git", "config", "remote.lhcb-nightlies.url"], cwd=proj_build_root
            )
            git_info["name"] = os.path.basename(os.path.splitext(url)[0])

            git_info["commit"] = check_output(
                ["git", "rev-parse", "HEAD"], cwd=proj_build_root
            ).strip()
            git_info["files"] = set(
                check_output(
                    ["git", "ls-tree", "--name-only", "-r", "HEAD"], cwd=proj_build_root
                ).splitlines()
            )
            git_info["url"] = (
                "https://gitlab.cern.ch/lhcb-nightlies/" "{name}/blob/{commit}/"
            ).format(**git_info)
    except CalledProcessError:
        # could not extract git info
        git_info = {}

    reports = {}
    legacy_build_logs = []

    for subdir, logs in build_logs.items():
        data = []
        for item in logs.values():
            for i in item:
                reports.setdefault(subdir, []).extend(scan_build_log(i["stderr"]))
                data.append(" ".join(i["cmd"]))
                if i["stdout"]:
                    data.append(i["stdout"])
                if i["stderr"]:
                    data.append(i["stderr"])
        legacy_build_logs.append((subdir, data))

    from .html_utils import ANSI2HTML, ClassifyByLineNo, TableizeLine, WrapLine

    html_section = {}
    for chunk_id, chunk in legacy_build_logs:
        classification = ClassifyByLineNo(
            (
                (issue.log_range[0] + 1, issue.log_range[1] + 1),
                {"warning": "alert-warning", "error": "alert-danger"}.get(
                    issue.severity, ""
                ),
            )
            for issue in reports[chunk_id]
        )
        actions = [
            lambda line: html.escape(line, quote=True),
            ANSI2HTML(),
            WrapLine("pre"),
            TableizeLine(
                line_id_prefix=chunk_id + "_",
                add_line_nos=True,
                row_class=classification,
            ),
        ]
        if git_info:
            actions.append(
                IssueLinker(
                    [
                        issue
                        for issue in reports[chunk_id]
                        if issue.source.name in git_info["files"]
                    ],
                    GitlabUrlFormat(git_info["url"]),
                )
            )

        def process(line):
            for action in actions:
                line = action(line)
            return line

        html_section[chunk_id] = TABLE_HEAD
        for line in (process(line.rstrip()) + "\n" for line in chunk):
            html_section[chunk_id] += line + "\n"
        html_section[chunk_id] += TABLE_TAIL

    jreport = report_dict(legacy_build_logs, reports)
    jreport["raw_logs"] = build_logs
    for section in jreport["sections"]:
        section["html"] = html_section[section["id"]]

    # filter issues to remove excluded ones
    jreport["ignored_issues"] = {severity: {} for severity in jreport["issues"]}
    exceptions = {
        severity: [(exp, re.compile(exp)) for exp in exceptions[severity]]
        for severity in exceptions or {}
    }
    for severity in exceptions:
        # remove excluded issues from the report and count them per regexp
        for text, exp in exceptions[severity]:
            before = len(jreport["issues"][severity])
            jreport["issues"][severity] = [
                issue
                for issue in jreport["issues"][severity]
                if not exp.search(issue["desc"])
            ]
            after = len(jreport["issues"][severity])
            jreport["ignored_issues"][severity][text] = before - after

    jreport.update(
        {
            "warnings": len(jreport["issues"]["warning"]),
            "errors": len(jreport["issues"]["error"]),
            "coverity_messages": len(jreport["issues"]["coverity"]),
        }
    )

    # limit the number of presented issues (see LBCORE-1210)
    # it's done here because I want jreport[severity's] to contain the total
    for severity in jreport["issues"]:
        count = len(jreport["issues"][severity])
        if count > MAX_REPORTED_ISSUES:
            jreport["ignored_issues"][severity]["too many issues"] = (
                count - MAX_REPORTED_ISSUES
            )
            jreport["issues"][severity][MAX_REPORTED_ISSUES:] = []

    # group repeated warnings
    issues = {"errors": {}, "warnings": {}, "coverity": {}}
    for (key, issue_type) in [
        ("errors", "error"),
        ("warnings", "warning"),
        ("coverity", "coverity"),
    ]:
        for issue in jreport["issues"][issue_type]:
            issues[key].setdefault(issue["desc"], {}).setdefault(
                issue["section_id"], []
            ).append(issue["anchor"])
    jreport["issues"] = {
        key: [{"desc": desc, "where": where} for desc, where in issues[key].items()]
        for key in issues
    }

    if extra_info:
        jreport.update(extra_info)

    return jreport



###############################################################################
# (c) Copyright 2020-2021 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "LICENSE".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
__version__ = "0.2.0"

import os

DATA_DIR = os.path.join(os.path.dirname(__file__), "data")
CMAKE_DIR = os.path.join(os.path.dirname(__file__), "data", "cmake")

from . import build, checkout



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['lb', 'lb.nightly.functions']

package_data = \
{'': ['*'],
 'lb.nightly.functions': ['data/*', 'data/cmake/*', 'data/cmake/modules/*']}

install_requires = \
['GitPython>=3.1.18,<4.0.0',
 'aiohttp>=3.7.4,<4.0.0',
 'archspec>=0.1.2,<0.2.0',
 'lb-nightly-configuration>=0.1.4,<0.2.0',
 'lb-nightly-db>=0.1,<0.2',
 'lb-nightly-utils>=0.2.0,<0.3.0',
 'lbplatformutils>=4.3.8,<5.0.0',
 'opensearch-py>=1.0.0,<2.0.0',
 'python-gitlab>=2.9.0,<3.0.0',
 'spython>=0.1.14,<0.2.0']

entry_points = \
{'console_scripts': ['lb-wrapcmd = lb.nightly.functions.common:lb_wrapcmd'],
 'lb.nightly.checkout_methods': ['git = lb.nightly.functions.checkout:git']}

setup_kwargs = {
    'name': 'lb-nightly-functions',
    'version': '0.2.0',
    'description': 'Core functions for LHCb Nightly and Continuous Integration Build System',
    'long_description': None,
    'author': 'Marco Clemencic',
    'author_email': 'marco.clemencic@cern.ch',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
